using IrController.Api.Filters;
using IrController.Api.LocalStorage;
using IrController.Api.Middleware;
using Microsoft.AspNetCore.Builder;
using Microsoft.AspNetCore.Hosting;
using Microsoft.Extensions.Configuration;
using Microsoft.Extensions.DependencyInjection;
using Microsoft.Extensions.Hosting;
using Microsoft.OpenApi.Models;
using Scrutor;
using System;
using System.Net;
using System.Net.Sockets;
using System.Threading.Tasks;

namespace IrController.Api
{
    public class Startup
    {
        public Startup(IConfiguration configuration)
        {
            Configuration = configuration;
        }

        public IConfiguration Configuration { get; }

        // This method gets called by the runtime. Use this method to add services to the container.
        public void ConfigureServices(IServiceCollection services)
        {
            services.AddSingleton<MyCache>();

            // Load Repositories with annotation 
            services.Scan(scan => scan
                  .FromAssemblyOf<Repository.CodeRepository>() // 1. Find the concrete classes
                    .AddClasses(x => x.WithAttribute<Repository.Attributes.LoadRepositoryAttribute>())//    to register
                      .UsingRegistrationStrategy(RegistrationStrategy.Skip) // 2. Define how to handle duplicates
                      .AsImplementedInterfaces()    // 2. Specify which services they are registered as
                      .WithSingletonLifetime()); // 3. Set the lifetime for the services

            services.AddControllers();

            services.AddSwaggerGen(c =>
            {
                c.OperationFilter<SwashbuckleOperationFilter>();
                c.SwaggerDoc("v1", new OpenApiInfo { Title = "IrController.Api", Version = "v1" });

                /// Replace . in namespace to AAAIIIIIAAA for ApiNamespaceGenerator
                c.CustomSchemaIds(x => x.FullName.Replace(".", "AAAIIIIIAAA"));


                // Token header
                c.AddSecurityDefinition("oauth2", new OpenApiSecurityScheme
                {
                    Description = "Authorization header",
                    In = ParameterLocation.Header,
                    Name = "Authorization",
                    Type = SecuritySchemeType.ApiKey
                });

                c.AddSecurityRequirement(new OpenApiSecurityRequirement
                {
                    {
                        new OpenApiSecurityScheme
                        {
                            Reference = new OpenApiReference { Type = ReferenceType.SecurityScheme, Id = "oauth2" }
                        },
                        new[] { "readAccess", "writeAccess" }
                    }
                });
            });
        }

        // This method gets called by the runtime. Use this method to configure the HTTP request pipeline.
        public void Configure(IApplicationBuilder app, IWebHostEnvironment env, MyCache cache)
        {
            if (env.IsDevelopment())
            {
                app.UseDeveloperExceptionPage();
                app.UseSwagger(c =>
                   c.SerializeAsV2 = true
                );
                app.UseSwaggerUI(c => c.SwaggerEndpoint("/swagger/v1/swagger.json", "IrController.Api v1"));
            }

            //app.UseHttpsRedirection();

            cache.LoadUserTokensCache();

            // Login
            app.UseAuthenticationMiddleware();

            // Logging errors
            app.UseExceptionMiddleware();

            // Authentication middleware
            //app.UseAuthenticationMiddleware();

            app.UseRouting();

            app.UseAuthorization();

            app.UseEndpoints(endpoints =>
            {
                endpoints.MapControllers();
            });

            Listen();
        }

        static async Task Listen()
        {
            var listener = new TcpListener(IPAddress.Any, 10000);
            listener.Start();
            Console.WriteLine("Waiting for a connection");
            var socket = await listener.AcceptSocketAsync();
            Console.WriteLine("Client connected!");
        }
    }
}
