﻿using IrController.DbModel.Db;
using IrController.DbModel.Db.Model;
using IrController.Model;
using IrController.Repository.Factories;
using IrController.Repository.Models;
using Microsoft.EntityFrameworkCore;
using System;
using System.Collections.Generic;
using System.Drawing;
using System.Linq;
using System.Net;

namespace IrController.Repository
{
    [Attributes.LoadRepository]
    public class CodeRepository : ICodeRepository
    {
        private readonly IIrControllerContextFactory _dbContextFactory;

        /// <summary>
        /// How many times will be send command to end device
        /// </summary>
        private const int REPETITION = 1;
        private ILogRepository _logRepository;

        public CodeRepository(IIrControllerContextFactory dbContextFactory, ILogRepository logRepository)
        {
            _dbContextFactory = dbContextFactory;
            _logRepository = logRepository;
        }

        public bool GetNewCode(string com, string add, string code)
        {
            using (IrControllerContext _db = _dbContextFactory.CreateDbContext())
            {
                try
                {
                    MissingCode found = _db.MissingCodes.FirstOrDefault(x => x.IrProtocol == code && x.SAddress == add && x.IrCommand == com);

                    if (found != null)
                        found.DateTimeCreated = DateTime.Now;
                    else
                    {
                        _db.MissingCodes.Add(new DbModel.Db.Model.MissingCode
                        {
                            IrProtocol = code,
                            SAddress = add,
                            IrCommand = com,
                            DateTimeCreated = DateTime.Now
                        });
                    }

                    _db.SaveChanges();
                    return true;

                }
                catch (Exception e)
                {
                    return false;
                }
            }
        }

        public bool ManageCode(CodeModel code)
        {
            using (IrControllerContext _db = _dbContextFactory.CreateDbContext())
            {
                Code dbCode = new Code();

                if (code.Id > 0)
                    dbCode = _db.Codes.First(x => x.Id == code.Id);

                dbCode.Name = code.Name;
                dbCode.IrProtocol = code.IrProtocol;
                dbCode.SAddress = code.Saddress;
                dbCode.IrCommand = code.IrCommand;
                dbCode.DeviceId = code.DeviceId;
                dbCode.DeviceFunctionId = code.FunctionId;
                dbCode.ColorValue = code.ColorValue;
                dbCode.BrightnessUp = code.BrightnessUp;
                CleanMissingCode(_db, code);

                if (code.Id < 1)
                    _db.Codes.Add(dbCode);

                try
                {
                    _db.SaveChanges();
                    return true;
                }
                catch (Exception)
                {
                    return false;
                }
            }
        }

        /// <summary>
        /// Remove assigned code
        /// </summary>
        /// <param name="irProtocol"></param>
        /// <param name="sAddress"></param>
        /// <param name="irCommand"></param>
        private void CleanMissingCode(IrControllerContext _db, CodeModel code)
        {
            _db.MissingCodes.RemoveRange(_db.MissingCodes.Where(x =>
            x.IrProtocol == code.IrProtocol &&
            x.SAddress == code.Saddress &&
            x.IrCommand == code.IrCommand
            ));
        }

        public bool WriteCommand(string com, string add, string irp)
        {
            using (IrControllerContext _db = _dbContextFactory.CreateDbContext())
            {
                MissingCode missingCode = _db.MissingCodes.FirstOrDefault(x => x.IrProtocol == irp && x.SAddress == add && x.IrCommand == com);

                if (missingCode == null)
                    _db.MissingCodes.Add(new MissingCode
                    {
                        IrCommand = com,
                        SAddress = add,
                        IrProtocol = irp,
                        DateTimeCreated = DateTime.Now
                    });
                else
                    missingCode.DateTimeCreated = DateTime.Now;

                _db.SaveChanges();

                return true;
            }
        }

        private bool SendCommand(SendModel sendData)
        {
            using (IrControllerContext _db = _dbContextFactory.CreateDbContext())
            {


                HttpWebRequest request = null;
                HttpWebResponse response = null;

                try
                {
                    request = (HttpWebRequest)WebRequest.Create($"http://{sendData.Ip}:50080/run?com:{sendData.Command}&add:{sendData.SAdress}&code:{sendData.IrProtocol}&rep={REPETITION}");
                    request.Method = "GET";

                    response = (HttpWebResponse)request.GetResponse();

                }
                catch (Exception e)
                {
                    _logRepository.AddLog(e.Message);
                }
                finally
                {
                    request = null;
                    response = null;
                }

                return true;
            }
        }

        public List<CodeModel> GetAllCodes()
        {
            using (IrControllerContext _db = _dbContextFactory.CreateDbContext())
            {
                return _db.Codes.Select(x => new CodeModel
                {
                    Id = x.Id,
                    Name = x.Name,
                    IrCommand = x.IrCommand,
                    IrProtocol = x.IrProtocol,
                    Saddress = x.SAddress,
                    DeviceId = x.DeviceId,
                    DeviceName = x.Device.Name,
                    FunctionId = x.DeviceFunctionId,
                    BrightnessUp = x.BrightnessUp,
                    ColorValue = x.ColorValue,
                    FunctionName = x.DeviceFunctionId != null ? x.DeviceFunction.Name : ""
                }).ToList();
            }
        }

        public List<CodeModel> GetCoddesByDevice(int deviceId)
        {
            using (IrControllerContext _db = _dbContextFactory.CreateDbContext())
            {
                return _db.Codes.Where(x => x.DeviceId == deviceId).Select(x => new CodeModel
                {
                    Id = x.Id,
                    Name = x.Name,
                    IrCommand = x.IrCommand,
                    IrProtocol = x.IrProtocol,
                    Saddress = x.SAddress,
                    DeviceId = x.DeviceId,
                    DeviceName = x.Device.Name,
                    FunctionId = x.DeviceFunctionId,
                    BrightnessUp = x.BrightnessUp,
                    ColorValue = x.ColorValue,
                    FunctionName = x.DeviceFunctionId != null ? x.DeviceFunction.Name : ""
                }).ToList();
            }
        }

        public List<DeviceFunctionModel> GetDeviceFunctions(int deviceId)
        {
            using (IrControllerContext _db = _dbContextFactory.CreateDbContext())
            {
                int deviceCategoryId = _db.Devices.First(x => x.Id == deviceId).DeviceTemplateId;
                return _db.Device_Templates_x_Functions.Where(x => x.DeviceTemplateId == deviceCategoryId).Select(x => new DeviceFunctionModel
                {
                    Id = x.DeviceFunction.Id,
                    Name = x.DeviceFunction.Name
                }).ToList();
            }
        }

        public CodeModel GetCode(int id)
        {
            using (IrControllerContext _db = _dbContextFactory.CreateDbContext())
            {
                return _db.Codes.Where(x => x.Id == id).Select(x => new CodeModel
                {
                    Id = x.Id,
                    Name = x.Name,
                    IrCommand = x.IrCommand,
                    IrProtocol = x.IrProtocol,
                    Saddress = x.SAddress,
                    DeviceId = x.DeviceId,
                    DeviceName = x.Device.Name,
                    FunctionId = x.DeviceFunctionId,
                    ColorValue = x.ColorValue,
                    BrightnessUp = x.BrightnessUp,
                    FunctionName = x.DeviceFunctionId != null ? x.DeviceFunction.Name : ""
                }).First();
            }
        }

        public CodeModel GetMissingCode(int id)
        {
            using (IrControllerContext _db = _dbContextFactory.CreateDbContext())
            {
                return _db.MissingCodes.Where(x => x.Id == id).Select(x => new CodeModel
                {
                    Id = x.Id,
                    IrCommand = x.IrCommand,
                    IrProtocol = x.IrProtocol,
                    Saddress = x.SAddress,
                    FunctionName = ""
                }).First();
            }
        }

        public List<CodeModel> GetAllMissingCodes()
        {
            using (IrControllerContext _db = _dbContextFactory.CreateDbContext())
            {
                return _db.MissingCodes.Select(x => new CodeModel
                {
                    Id = x.Id,
                    IrCommand = x.IrCommand,
                    IrProtocol = x.IrProtocol,
                    Saddress = x.SAddress,
                    FunctionName = "",
                    Created = x.DateTimeCreated
                }).OrderByDescending(x => x.Created).ToList();
            }
        }

        public bool ManageMissingCode(CodeModel code)
        {
            using (IrControllerContext _db = _dbContextFactory.CreateDbContext())
            {
                Code dbCode = new Code { };
                dbCode.Name = code.Name;
                dbCode.IrProtocol = code.IrProtocol;
                dbCode.SAddress = code.Saddress;
                dbCode.IrCommand = code.IrCommand;
                dbCode.DeviceId = code.DeviceId;
                dbCode.DeviceFunctionId = code.FunctionId;

                _db.Codes.Add(dbCode);

                try
                {
                    
                    _db.MissingCodes.Remove(_db.MissingCodes.First(x => x.Id == code.Id));
                    _db.SaveChanges();
                    return true;
                }
                catch (Exception)
                {
                    return false;
                }
            }
        }

        public string RemoveCode(int id)
        {
            using (IrControllerContext _db = _dbContextFactory.CreateDbContext())
            {
                try
                {
                    _db.Codes.Remove(_db.Codes.First(x => x.Id == id));
                    _db.SaveChanges();
                }
                catch (Exception e)
                {
                    return e.Message;
                }

                return "";

            }
        }

        public string RemoveMissingCode(int id)
        {
            using (IrControllerContext _db = _dbContextFactory.CreateDbContext())
            {
                try
                {
                    _db.MissingCodes.Remove(_db.MissingCodes.First(x => x.Id == id));
                    _db.SaveChanges();
                }
                catch (Exception e)
                {
                    return e.Message;
                }

                return "";

            }
        }

        public string GetStatus(int codeId)
        {
            using (IrControllerContext _db = _dbContextFactory.CreateDbContext())
            {
                var code = _db.Codes.First(x => x.Id == codeId);
                if (string.IsNullOrWhiteSpace(code.Status))
                    return "0";

                return code.Status;
            }
        }

        public bool SetSwitch(int value, int codeId)
        {
            using (IrControllerContext _db = _dbContextFactory.CreateDbContext())
            {
                SendModel sendData = _db.Codes.Where(x => x.Id == codeId).Select(x => new SendModel
                {
                    Ip = x.Device.Arduino.Ip,
                    SAdress = x.SAddress,
                    Command = x.IrCommand,
                    IrProtocol = x.IrProtocol
                }).First();

                if (!SendCommand(sendData))
                    return false;



                var code = _db.Codes.Include(x => x.DeviceFunction).First(x => x.Id == codeId);
                var category = _db.DeviceFunctions.Include(x => x.DeviceCategory).First(x => x.Id == code.DeviceFunctionId).DeviceCategory;
                var allFunctions = _db.DeviceFunctions.Where(x => x.DeviceCategoryId == category.Id).Select(x => x.Id);

                foreach (var dbCode in _db.Codes.Where(x => x.DeviceId == code.DeviceId && allFunctions.Any(y => y == x.DeviceFunctionId)).ToList())
                {
                    dbCode.Status = value.ToString();
                }

                _db.SaveChanges();
                return true;
            }
        }

        public bool SetColor(string  value, int codeId)
        {
            using (IrControllerContext _db = _dbContextFactory.CreateDbContext())
            {
                var code = _db.Codes.Include(x => x.DeviceFunction).Include(x => x.Device.Arduino).First(x => x.Id == codeId);
                string ip = code.Device.Arduino.Ip;
                var codes = _db.Codes.Where(x => x.DeviceFunctionId == code.DeviceFunctionId && 
                    x.DeviceId == code.DeviceId &&
                    x.ColorValue.Length == 6
                    ).ToList();
                code = GetClosest(codes, value.ToString());

                SendModel sendData = new SendModel
                {
                    Ip = ip,
                    SAdress = code.SAddress,
                    Command = code.IrCommand,
                    IrProtocol = code.IrProtocol
                };

                if (!SendCommand(sendData))
                    return false;


                
                var category = _db.DeviceFunctions.Include(x => x.DeviceCategory).First(x => x.Id == code.DeviceFunctionId).DeviceCategory;
                var allFunctions = _db.DeviceFunctions.Where(x => x.DeviceCategoryId == category.Id).Select(x => x.Id);

                foreach (var dbCode in _db.Codes.Where(x => x.DeviceId == code.DeviceId && allFunctions.Any(y => y == x.DeviceFunctionId)).ToList())
                {
                    dbCode.Status = value.ToString();
                }

                _db.SaveChanges();
                return true;
            }
        }

        /// <summary>
        /// Get closest color
        /// </summary>
        /// <param name="codes"></param>
        /// <param name="current"></param>
        /// <returns></returns>
        private Code GetClosest(List<Code> codes, string current)
        {
            List<int> tempColor = GetChunks(current);
            Color final = Color.FromArgb(tempColor[0], tempColor[1], tempColor[2]);
            Code result = codes.First();
            tempColor = GetChunks(result.ColorValue);
            int bestResult = GetDistance(final, Color.FromArgb(tempColor[0], tempColor[1], tempColor[2]));
            
            foreach (var code in codes)
            {
                tempColor = GetChunks(code.ColorValue);
                if (GetDistance(final, Color.FromArgb(tempColor[0], tempColor[1], tempColor[2])) < bestResult)
                    result = code;
            }
            return result;
        }

        /// <summary>
        /// Chunk string to RGB
        /// </summary>
        /// <param name="value"></param>
        /// <returns></returns>
        private List<int> GetChunks(string value)
        {
            List<int> triplets = new List<int>();
            while (value.Length > 2)
            {
                triplets.Add(int.Parse(value.Substring(0, 2), System.Globalization.NumberStyles.HexNumber));
                value = value.Substring(2);
            }
            if (value != "")
                triplets.Add(int.Parse(value, System.Globalization.NumberStyles.HexNumber));
            return triplets;
        }

        /// <summary>
        /// Get color distance
        /// </summary>
        /// <param name="current"></param>
        /// <param name="match"></param>
        /// <returns></returns>
        public int GetDistance(Color current, Color match)
        {
            int redDifference;
            int greenDifference;
            int blueDifference;

            redDifference = current.R - match.R;
            greenDifference = current.G - match.G;
            blueDifference = current.B - match.B;

            return redDifference * redDifference + greenDifference * greenDifference + blueDifference * blueDifference;
        }

        /// <summary>
        /// Count steps for brightness
        /// If below zero it means that is need to be darker.
        /// </summary>
        /// <param name="deviceSteps"></param>
        /// <param name="brightness"></param>
        /// <param name="current"></param>
        /// <returns></returns>
        private int BrightnessSteps(int? deviceSteps, int brightness, string current)
        {
            int toReturn = 0;
            if (deviceSteps == null || deviceSteps < 1)
            {
                deviceSteps = 3;
            }

            int currentBrightness = 0;
            bool success =  int.TryParse(current, out currentBrightness);

            int oneStep = 100/(int)deviceSteps;

            int difference = brightness - currentBrightness;

            // make it Lighter
            if (difference > 0)
            {
                for (int i = 1; i < deviceSteps; i++)
                {
                    if (Math.Abs(oneStep + currentBrightness - brightness) <= Math.Abs(currentBrightness - brightness) && Math.Abs(currentBrightness - oneStep) <= 100)
                    {
                        currentBrightness += oneStep;
                        toReturn++;
                    }
                        
                    else
                        return toReturn;
                }

            }
            else
            {
                //Darker
                for (int i = 1; i < deviceSteps; i++)
                {
                    if (Math.Abs(currentBrightness - oneStep - brightness) <= Math.Abs(currentBrightness - brightness) && Math.Abs(currentBrightness - oneStep) <= 100)
                    {
                        currentBrightness -= oneStep;
                        toReturn--;
                    }
                    else
                        return toReturn;
                }
            }
            

            return toReturn;
        }

        public bool SetBrightness(int value, int codeId)
        {
            using (IrControllerContext _db = _dbContextFactory.CreateDbContext())
            {
                var code = _db.Codes.Include(x => x.DeviceFunction).First(x => x.Id == codeId);
                int brSteps = BrightnessSteps(_db.Devices.First(x => x.Id == code.DeviceId).MaxBrightnessStep, value, code.Status);

                if (brSteps == 0)
                    return true;

                if (brSteps > 0)
                {
                    SendModel sendData = _db.Codes.Where(x => x.DeviceId == code.DeviceId && x.DeviceFunctionId == code.DeviceFunctionId && x.BrightnessUp).Select(x => new SendModel
                    {
                        Ip = x.Device.Arduino.Ip,
                        SAdress = x.SAddress,
                        Command = x.IrCommand,
                        IrProtocol = x.IrProtocol
                    }).First();

                    while (brSteps > 0)
                    {
                        if (!SendCommand(sendData))
                            return false;

                        brSteps--;
                    }
                }

                if (brSteps < 0)
                {
                    SendModel sendData = _db.Codes.Where(x => x.DeviceId == code.DeviceId && x.DeviceFunctionId == code.DeviceFunctionId && !x.BrightnessUp).Select(x => new SendModel
                    {
                        Ip = x.Device.Arduino.Ip,
                        SAdress = x.SAddress,
                        Command = x.IrCommand,
                        IrProtocol = x.IrProtocol
                    }).First();

                    while (brSteps < 0)
                    {
                        if (!SendCommand(sendData))
                            return false;

                        brSteps++;
                    }
                }

                var category = _db.DeviceFunctions.Include(x => x.DeviceCategory).First(x => x.Id == code.DeviceFunctionId).DeviceCategory;
                var allFunctions = _db.DeviceFunctions.Where(x => x.DeviceCategoryId == category.Id).Select(x => x.Id);

                foreach (var dbCode in _db.Codes.Where(x => x.DeviceId == code.DeviceId && allFunctions.Any(y => y == x.DeviceFunctionId)).ToList())
                {
                    dbCode.Status = value.ToString();
                }

                _db.SaveChanges();
                return true;
            }

        }
    }
}
