﻿using IrController.DbModel.Db;
using IrController.DbModel.Db.Model;
using IrController.Model;
using IrController.Repository.Factories;
using IrController.Repository.Helpers;
using IrController.Repository.Models;
using Microsoft.EntityFrameworkCore;
using Newtonsoft.Json;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Reflection;
using System.Text.Json;

namespace IrController.Repository
{
    [Attributes.LoadRepository]
    public class DeviceRepository : IDeviceRepository
    {
        private readonly IIrControllerContextFactory _dbContextFactory;
        public DeviceRepository(IIrControllerContextFactory dbContextFactory)
        {
            _dbContextFactory = dbContextFactory;
        }

        public bool ManageDevice(DeviceModel device)
        {
            using (IrControllerContext _db = _dbContextFactory.CreateDbContext())
            {
                List<Code> dbCodes = _db.Codes.Where(x => device.Id == device.Id).ToList();

                Device dbDevice = new Device();
                if (device.Id > 0)
                {
                    dbDevice = _db.Devices.Include(x => x.Codes).First(x => x.Id == device.Id);                   
                }
                else
                {
                    _db.Devices.Add(dbDevice);
                }

                dbDevice.Name = device.Name;
                dbDevice.MaxBrightnessStep = device.MaxBrightnessStep;
                dbDevice.DeviceTemplateId = device.DeviceTemplateId;

                try
                {
                    _db.SaveChanges();
                    return true;
                }
                catch (Exception e)
                {
                    return false;
                }

            }
        }

        public List<DeviceModel> GetDevices()
        {
            using (IrControllerContext _db = _dbContextFactory.CreateDbContext())
            {
                return _db.Devices.Select(x => new DeviceModel { 
                    Id = x.Id,
                    Name = x.Name,
                    DeviceTemplateId = x.DeviceTemplateId,
                    DeviceTemplateName = x.DeviceTemplate.Name,
                    DeviceArduinoId = x.ArduinoId,
                    DeviceArduinoName = x.ArduinoId != null ? x.Arduino.Name : "",
                    CodesSum = x.Codes.Count(),
                    MaxBrightnessStep = x.MaxBrightnessStep
                }).ToList();
            }
        }

        public string RemoveDevice(int id)
        {
            using (IrControllerContext _db = _dbContextFactory.CreateDbContext())
            {
                int countCodes = 0;
                try
                {
                    countCodes = _db.Devices.First(x => x.Id == id).Codes.Count;
                }
                catch (Exception)
                {
                    return "Device not found!";
                }

                if (countCodes > 1)
                    return "Device contain Codes, please remove codes first";

                try
                {
                    _db.Devices.Remove(_db.Devices.First(x => x.Id == id));
                }
                catch (Exception e)
                {
                    return e.Message;
                }

                return "";
                
            }
        }

        public DeviceModel GetDevice(int id)
        {
            using (IrControllerContext _db = _dbContextFactory.CreateDbContext())
            {
                return _db.Devices.Select(x => new DeviceModel
                {
                    Id = x.Id,
                    Name = x.Name,
                    DeviceTemplateId = x.DeviceTemplateId,
                    DeviceTemplateName = x.DeviceTemplate.Name,
                    DeviceArduinoId = x.ArduinoId,
                    MaxBrightnessStep = x.MaxBrightnessStep,
                    DeviceArduinoName = x.ArduinoId != null ? x.Arduino.Name : "",
                }).First(x => x.Id == id);
            }
        }

        public List<DeviceTemplateModel> GetDeviceTemplates()
        {
            using (IrControllerContext _db = _dbContextFactory.CreateDbContext())
            {
                return _db.DeviceTemplates.Select(x => new DeviceTemplateModel
                {
                    Id = x.Id,
                    Name = x.Name,
                }).ToList();
            }
        }

        public JsonReponse GetHttpRgbPushAccessory(string apiUrl, int id)
        {
            using (IrControllerContext _db = _dbContextFactory.CreateDbContext())
            {
                JsonReponse result = new JsonReponse();

                Models.Addons.HttpRgbPushModel.Accessory accessory = _db.Devices.Where(x => x.Id == id).Select(x => new Models.Addons.HttpRgbPushModel.Accessory
                {
                    Name = x.Name,
                    AccessoryString = x.DeviceTemplate.Accessory,
                    Service = x.DeviceTemplate.Service
                }).First();

                int deviceTemplate = _db.Devices.First(x => x.Id == id).DeviceTemplateId;

                var codes = _db.Codes.Where(x => x.DeviceId == id && x.DeviceFunctionId != null).Select(x => new 
                {
                    FunctionName = x.DeviceFunction.CodeString,
                    CategoryName = x.DeviceFunction.DeviceCategory.Name,
                    Url = x.DeviceFunction.Url,
                    Id = x.Id
                }).ToList();


                foreach (var code in codes)
                {
                    PropertyInfo propertyInfo = accessory.GetType().GetProperty(code.CategoryName);

                    if (propertyInfo.GetValue(accessory, null) == null)
                    {
                        var newObject = Activator.CreateInstance(Type.GetType(propertyInfo.PropertyType.FullName), apiUrl + "api/Code/GetStatus?codeId=" + code.Id);
                        propertyInfo.SetValue(accessory, newObject, null);
                    }
                        

                    var settingObject = propertyInfo.GetValue(accessory, null);
                    propertyInfo = propertyInfo.GetValue(accessory, null).GetType().GetProperty(code.FunctionName);
                    propertyInfo.SetValue(settingObject, apiUrl + "api" + code.Url + "&codeId=" + code.Id.ToString(), null);
                }

                if (accessory.Switch == null)
                    result.Warning = "ON and OFF is not defined!";
                else if(string.IsNullOrWhiteSpace(accessory.Switch.powerOn))
                    result.Warning = "ON  is not defined!";
                else if (string.IsNullOrWhiteSpace(accessory.Switch.powerOff))
                    result.Warning = "OFF is not defined!";

                result.Object = JsonConvert.SerializeObject(accessory, Formatting.Indented);
                return result;
            }
        }

        public JsonReponse GetArduinoSender(string apiUrl, int id)
        {
            using (IrControllerContext _db = _dbContextFactory.CreateDbContext())
            {
                Arduino arduino = _db.Devices.Where(x => x.Id == id).Select(x => x.Arduino).First();

                return new JsonReponse { Object = ArduinoHelper.Sender(arduino.Mac, arduino.Ip, arduino.Token) };
            }
        }

        public JsonReponse GetArduinoReceiver(string apiUrl, int id)
        {
            using (IrControllerContext _db = _dbContextFactory.CreateDbContext())
            {
                Arduino arduino = _db.Devices.Where(x => x.Id == id).Select(x => x.Arduino).First();


                return new JsonReponse { Object = ArduinoHelper.Receiver(arduino.Mac, arduino.Ip, apiUrl, arduino.Token) } ;
            }
        }
    }
}
